import asyncio
import functools

import aiohttp
from bs4 import BeautifulSoup

from .. import fhem, generic, utils


class fhem_forum(generic.FhemModule):

    URL_UNREADREPLIES = "https://forum.fhem.de/index.php?action=unreadreplies"
    URL_UNREAD = "https://forum.fhem.de/index.php?action=unread"

    def __init__(self, logger):
        super().__init__(logger)

        attr_config = {
            "interval": {
                "default": 15,
                "format": "int",
                "help": "Change interval in minutes, default is 15.",
            },
            "max_topics": {
                "default": 10,
                "format": "int",
                "help": "Maximum number of topics to show in readings.",
            },
            "keywords_unread_replies": {
                "default": "",
                "help": "Only topics with this keywords will be listed.",
                "function": "set_attr_keywords",
            },
            "keywords_unread": {
                "default": "",
                "help": "Only topics with this keywords will be listed.",
                "function": "set_attr_keywords",
            },
        }
        self.set_attr_config(attr_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define fhem.forum fhempy fhem_forum FHEM_COOKIE"

        self.cookie = args[3]
        self.first_run = True

        self.create_async_task(self.update_loop())

    async def set_attr_keywords(self, hash):
        self.create_async_task(self.update_once())

    async def update_once(self):
        stateset_ur = await self.get_fhem_data(fhem_forum.URL_UNREADREPLIES)
        stateset_u = await self.get_fhem_data(fhem_forum.URL_UNREAD)
        if stateset_ur is False and stateset_u is False:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "-", 1)

    async def update_loop(self):
        while True:
            # aiohttp get
            await self.update_once()
            await asyncio.sleep(self._attr_interval * 60)

    async def get_fhem_data(self, url):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    url,
                    headers={"cookie": "FHEM-Forum588=" + self.cookie},
                ) as resp:
                    if resp.status == 200:
                        return await self.handle_response(await resp.text(), url)
                    else:
                        await fhem.readingsSingleUpdate(
                            self.hash,
                            "state",
                            f"failed: HTTP error {resp.status}",
                            1,
                        )
                        self.logger.error(
                            f"Failed to fetch {self.update_url}, "
                            f"failed with status {resp.status}"
                        )
        except Exception:
            self.logger.exception("Failed to update")

    def contains_keyword(self, topic, keywords):
        if keywords == "":
            return True

        for kw in keywords.split(","):
            if topic.find(kw) > -1:
                return True
        return False

    def get_soup_entries(self, response):
        soup = BeautifulSoup(response, "html.parser")
        tbody = soup.find("tbody")
        if tbody is None:
            return []
        entries = tbody.findAll("tr")
        return entries

    async def handle_response(self, response, url):
        # bs4
        stateset = False
        await fhem.readingsBeginUpdate(self.hash)
        try:
            reading = "unread_replies"
            keywords = self._attr_keywords_unread_replies
            if url == fhem_forum.URL_UNREAD:
                reading = "unread"
                keywords = self._attr_keywords_unread

            entries = await utils.run_blocking(
                functools.partial(self.get_soup_entries, response)
            )

            i = 1
            for entry in entries:
                subject = entry.find("td", {"class": "subject windowbg2"})
                if subject is None:
                    continue

                if not self.contains_keyword(subject.find("span").text, keywords):
                    continue

                link_to_new = subject.findAll("a")[1]
                link_to_new.next_element.replace_with(subject.find("span").text)
                link_to_new.attrs["target"] = "_blank"

                last_post = " ".join(
                    entry.find("td", {"class": "lastpost windowbg2"}).text.split()
                )

                ret = await fhem.readingsBulkUpdateIfChanged(
                    self.hash,
                    f"topic_{reading}_{i:02d}",
                    f"<html>{link_to_new}<br>{last_post}</html>",
                )

                if i == 1 and (ret is not None or self.first_run is True):
                    self.first_run = False
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        "state",
                        f"<html>{link_to_new}<br>{last_post}</html>",
                    )

                stateset = True

                i += 1
                if i > self._attr_max_topics:
                    break

            while i <= self._attr_max_topics:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, f"topic_{reading}_{i:02d}", "-"
                )
                i += 1

        except Exception as ex:
            self.logger.exception(f"Failed to handle data from {url}")
            await fhem.readingsBulkUpdate(self.hash, "state", {ex}, 1)

        await fhem.readingsEndUpdate(self.hash, 1)

        return stateset



import logging
import random
import secrets
import string
import time
from datetime import datetime, timedelta

import pytz
import requests
from requests import RequestException, Response
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.ssl_ import create_urllib3_context

from .const import DATE_FORMAT, DOMAIN
from .KiaUvoApiImpl import KiaUvoApiImpl
from .Token import Token

_LOGGER = logging.getLogger(__name__)

CIPHERS = "DEFAULT@SECLEVEL=1"


class cipherAdapter(HTTPAdapter):
    """
    A HTTPAdapter that re-enables poor ciphers required by Hyundai.
    """

    def init_poolmanager(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs["ssl_context"] = context
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs["ssl_context"] = context
        return super().proxy_manager_for(*args, **kwargs)


class AuthError(RequestException):
    pass


def request_with_active_session(func):
    def request_with_active_session_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AuthError:
            _LOGGER.debug(f"got invalid session, attempting to repair and resend")
            self = args[0]
            token = kwargs["token"]
            new_token = self.login()
            _LOGGER.debug(
                f"old token:{token.access_token}, new token:{new_token.access_token}"
            )
            token.access_token = new_token.access_token
            token.vehicle_regid = new_token.vehicle_regid
            token.valid_until = new_token.valid_until
            json_body = kwargs.get("json_body", None)
            if json_body is not None and json_body.get("vinKey", None):
                json_body["vinKey"] = [token.vehicle_regid]
            response = func(*args, **kwargs)
            return response

    return request_with_active_session_wrapper


def request_with_logging(func):
    def request_with_logging_wrapper(*args, **kwargs):
        url = kwargs["url"]
        json_body = kwargs.get("json_body")
        if json_body is not None:
            _LOGGER.debug(f"sending {url} request with {json_body}")
        else:
            _LOGGER.debug(f"sending {url} request")
        response = func(*args, **kwargs)
        _LOGGER.debug(f"got response {response.text}")
        response_json = response.json()
        if response_json["status"]["statusCode"] == 0:
            return response
        if (
            response_json["status"]["statusCode"] == 1
            and response_json["status"]["errorType"] == 1
            and response_json["status"]["errorCode"] == 1003
        ):
            _LOGGER.debug(f"error: session invalid")
            raise AuthError
        _LOGGER.error(f"error: unknown error response {response.text}")
        raise RequestException

    return request_with_logging_wrapper


class KiaUvoAPIUSA(KiaUvoApiImpl):
    def __init__(
        self,
        username: str,
        password: str,
        region: int,
        brand: int,
        use_email_with_geocode_api: bool = False,
        pin: str = "",
    ):
        super().__init__(
            username, password, region, brand, use_email_with_geocode_api, pin
        )
        self.last_action_tracked = True
        self.last_action_xid = None
        self.last_action_completed = False

        self.supports_soc_range = False

        # Randomly generate a plausible device id on startup
        self.device_id = (
            "".join(
                random.choice(string.ascii_letters + string.digits) for _ in range(22)
            )
            + ":"
            + secrets.token_urlsafe(105)
        )

        self.BASE_URL: str = "api.owners.kia.com"
        self.API_URL: str = "https://" + self.BASE_URL + "/apigw/v1/"
        self.sessions = requests.Session()
        self.sessions.mount(self.API_URL, cipherAdapter())

    def api_headers(self) -> dict:
        offset = time.localtime().tm_gmtoff / 60 / 60
        headers = {
            "content-type": "application/json;charset=UTF-8",
            "accept": "application/json, text/plain, */*",
            "accept-encoding": "gzip, deflate, br",
            "accept-language": "en-US,en;q=0.9",
            "apptype": "L",
            "appversion": "4.10.0",
            "clientid": "MWAMOBILE",
            "from": "SPA",
            "host": self.BASE_URL,
            "language": "0",
            "offset": str(int(offset)),
            "ostype": "Android",
            "osversion": "11",
            "secretkey": "98er-w34rf-ibf3-3f6h",
            "to": "APIGW",
            "tokentype": "G",
            "user-agent": "okhttp/3.12.1",
        }
        # should produce something like "Mon, 18 Oct 2021 07:06:26 GMT". May require adjusting locale to en_US
        date = datetime.now(tz=pytz.utc).strftime("%a, %d %b %Y %H:%M:%S GMT")
        headers["date"] = date
        headers["deviceid"] = self.device_id
        return headers

    def authed_api_headers(self, token: Token):
        headers = self.api_headers()
        headers["sid"] = token.access_token
        headers["vinkey"] = token.vehicle_regid
        return headers

    @request_with_active_session
    @request_with_logging
    def post_request_with_logging_and_active_session(
        self, token: Token, url: str, json_body: dict
    ) -> Response:
        headers = self.authed_api_headers(token)
        return self.sessions.post(url, json=json_body, headers=headers)

    @request_with_active_session
    @request_with_logging
    def get_request_with_logging_and_active_session(
        self, token: Token, url: str
    ) -> Response:
        headers = self.authed_api_headers(token)
        return self.sessions.get(url, headers=headers)

    def login(self) -> Token:
        username = self.username
        password = self.password

        # Sign In with Email and Password and Get Authorization Code

        url = self.API_URL + "prof/authUser"

        data = {
            "deviceKey": "",
            "deviceType": 2,
            "userCredential": {"userId": username, "password": password},
        }
        headers = self.api_headers()
        response = self.sessions.post(url, json=data, headers=headers)
        _LOGGER.debug(f"{DOMAIN} - Sign In Response {response.text}")
        session_id = response.headers.get("sid")
        if not session_id:
            raise Exception(
                f"no session id returned in login. Response: {response.text} headers {response.headers} cookies {response.cookies}"
            )
        _LOGGER.debug(f"got session id {session_id}")

        # Get Vehicles
        url = self.API_URL + "ownr/gvl"
        headers = self.api_headers()
        headers["sid"] = session_id
        response = self.sessions.get(url, headers=headers)
        _LOGGER.debug(f"{DOMAIN} - Get Vehicles Response {response.text}")
        response = response.json()
        vehicle_summary = response["payload"]["vehicleSummary"][0]
        vehicle_name = vehicle_summary["nickName"]
        vehicle_id = vehicle_summary["vehicleIdentifier"]
        vehicle_vin = vehicle_summary["vin"]
        vehicle_key = vehicle_summary["vehicleKey"]
        vehicle_model = vehicle_summary["modelName"]
        vehicle_registration_date = vehicle_summary.get("enrollmentDate", "missing")

        valid_until = (datetime.now() + timedelta(hours=1)).strftime(DATE_FORMAT)

        # using vehicle_VIN as device ID
        # using vehicle_key as vehicle_regid

        token = Token({})
        token.set(
            session_id,
            None,
            vehicle_vin,
            vehicle_name,
            vehicle_id,
            vehicle_key,
            vehicle_model,
            vehicle_registration_date,
            valid_until,
            "NoStamp",
        )

        return token

    def get_cached_vehicle_status(self, token: Token):
        url = self.API_URL + "cmm/gvi"

        body = {
            "vehicleConfigReq": {
                "airTempRange": "0",
                "maintenance": "1",
                "seatHeatCoolOption": "0",
                "vehicle": "1",
                "vehicleFeature": "0",
            },
            "vehicleInfoReq": {
                "drivingActivty": "0",
                "dtc": "1",
                "enrollment": "1",
                "functionalCards": "0",
                "location": "1",
                "vehicleStatus": "1",
                "weather": "0",
            },
            "vinKey": [token.vehicle_regid],
        }
        response = self.post_request_with_logging_and_active_session(
            token=token, url=url, json_body=body
        )

        response_body = response.json()
        vehicle_status = response_body["payload"]["vehicleInfoList"][0][
            "lastVehicleInfo"
        ]["vehicleStatusRpt"]["vehicleStatus"]

        vehicle_status["time"] = vehicle_status["syncDate"]["utc"]

        if vehicle_status["batteryStatus"].get("stateOfCharge"):
            vehicle_status["battery"] = {
                "batSoc": vehicle_status["batteryStatus"]["stateOfCharge"],
            }

        if vehicle_status.get("evStatus"):
            vehicle_status["evStatus"]["remainTime2"] = {
                "atc": vehicle_status["evStatus"]["remainChargeTime"][0]["timeInterval"]
            }

        vehicle_status["doorOpen"] = vehicle_status["doorStatus"]
        vehicle_status["trunkOpen"] = vehicle_status["doorStatus"]["trunk"]
        vehicle_status["hoodOpen"] = vehicle_status["doorStatus"]["hood"]

        if vehicle_status.get("tirePressure"):
            vehicle_status["tirePressureLamp"] = {
                "tirePressureLampAll": vehicle_status["tirePressure"]["all"]
            }

        climate_data = vehicle_status["climate"]
        vehicle_status["airCtrlOn"] = climate_data["airCtrl"]
        vehicle_status["defrost"] = climate_data["defrost"]
        vehicle_status["sideBackWindowHeat"] = climate_data["heatingAccessory"][
            "rearWindow"
        ]
        vehicle_status["sideMirrorHeat"] = climate_data["heatingAccessory"][
            "sideMirror"
        ]
        vehicle_status["steerWheelHeat"] = climate_data["heatingAccessory"][
            "steeringWheel"
        ]

        vehicle_status["airTemp"] = climate_data["airTemp"]
        vehicle_status["dte"] = vehicle_status.get("distanceToEmpty", None)
        vehicle_data = {
            "vehicleStatus": vehicle_status,
            "odometer": {
                "value": float(
                    response_body["payload"]["vehicleInfoList"][0]["vehicleConfig"][
                        "vehicleDetail"
                    ]["vehicle"]["mileage"]
                ),
                "unit": 3,
            },
            "nextService": {
                "value": float(
                    response_body["payload"]["vehicleInfoList"][0]["vehicleConfig"][
                        "maintenance"
                    ]["nextServiceMile"]
                ),
                "unit": 3,
            },
            "vehicleLocation": response_body["payload"]["vehicleInfoList"][0][
                "lastVehicleInfo"
            ]["location"],
        }
        return vehicle_data

    def get_location(self, token: Token):
        pass

    def get_pin_token(self, token: Token):
        pass

    def update_vehicle_status(self, token: Token):
        url = self.API_URL + "rems/rvs"
        body = {
            "requestType": 0  # value of 1 would return cached results instead of forcing update
        }
        self.post_request_with_logging_and_active_session(
            token=token, url=url, json_body=body
        )

    def check_last_action_status(self, token: Token):
        url = self.API_URL + "cmm/gts"
        body = {"xid": self.last_action_xid}
        response = self.post_request_with_logging_and_active_session(
            token=token, url=url, json_body=body
        )
        response_json = response.json()
        self.last_action_completed = all(
            v == 0 for v in response_json["payload"].values()
        )
        return self.last_action_completed

    def lock_action(self, token: Token, action):
        _LOGGER.debug(f"Action for lock is: {action}")
        if action == "close":
            url = self.API_URL + "rems/door/lock"
            _LOGGER.debug(f"Calling Lock")
        else:
            url = self.API_URL + "rems/door/unlock"
            _LOGGER.debug(f"Calling unlock")

        response = self.get_request_with_logging_and_active_session(
            token=token, url=url
        )

        self.last_action_xid = response.headers["Xid"]

    def start_climate(
        self, token: Token, set_temp, duration, defrost, climate, heating
    ):
        url = self.API_URL + "rems/start"
        if set_temp < 62:
            set_temp = "LOW"
        elif set_temp > 82:
            set_temp = "HIGH"
        body = {
            "remoteClimate": {
                "airCtrl": climate,
                "airTemp": {
                    "unit": 1,
                    "value": str(set_temp),
                },
                "defrost": defrost,
                "heatingAccessory": {
                    "rearWindow": int(heating),
                    "sideMirror": int(heating),
                    "steeringWheel": int(heating),
                },
                "ignitionOnDuration": {
                    "unit": 4,
                    "value": duration,
                },
            }
        }
        response = self.post_request_with_logging_and_active_session(
            token=token, url=url, json_body=body
        )
        self.last_action_xid = response.headers["Xid"]

    def stop_climate(self, token: Token):
        url = self.API_URL + "rems/stop"
        response = self.get_request_with_logging_and_active_session(
            token=token, url=url
        )
        self.last_action_xid = response.headers["Xid"]

    def start_charge(self, token: Token):
        url = self.API_URL + "evc/charge"
        body = {"chargeRatio": 100}
        response = self.post_request_with_logging_and_active_session(
            token=token, url=url, json_body=body
        )
        self.last_action_xid = response.headers["Xid"]

    def stop_charge(self, token: Token):
        url = self.API_URL + "evc/cancel"
        response = self.get_request_with_logging_and_active_session(
            token=token, url=url
        )
        self.last_action_xid = response.headers["Xid"]

    def set_charge_limits(self, token: Token, ac_limit: int, dc_limit: int):
        url = self.API_URL + "evc/sts"
        body = {
            "targetSOClist": [
                {
                    "plugType": 0,
                    "targetSOClevel": dc_limit,
                },
                {
                    "plugType": 1,
                    "targetSOClevel": ac_limit,
                },
            ]
        }
        response = self.post_request_with_logging_and_active_session(
            token=token, url=url, json_body=body
        )
        self.last_action_xid = response.headers["Xid"]



from .const import BRANDS, DATA_VEHICLE_INSTANCE, DOMAIN, TOPIC_UPDATE
from .Vehicle import Vehicle


class KiaUvoEntity:
    def __init__(self, hass, config_entry, vehicle: Vehicle):
        self.hass = hass
        self.config_entry = config_entry
        self.vehicle = vehicle
        self.topic_update = TOPIC_UPDATE.format(vehicle.id)
        self.topic_update_listener = None

    async def async_added_to_hass(self):
        @callback
        def update():
            self.update_from_latest_data()
            self.async_write_ha_state()

        await super().async_added_to_hass()
        self.topic_update_listener = async_dispatcher_connect(
            self.hass, self.topic_update, update
        )
        self.async_on_remove(self.topic_update_listener)
        self.update_from_latest_data()

    @property
    def available(self) -> bool:
        return not not self.vehicle

    @property
    def should_poll(self) -> bool:
        return False

    @property
    def device_info(self):
        return {
            "identifiers": {(DOMAIN, self.vehicle.id)},
            "name": self.vehicle.name,
            "manufacturer": BRANDS[self.vehicle.kia_uvo_api.brand],
            "model": self.vehicle.model,
            "engine_type": self.vehicle.engine_type,
            "sw_version": self.vehicle.registration_date,
            "via_device": (DOMAIN, self.vehicle.id),
        }

    @callback
    def update_from_latest_data(self):
        self.vehicle = self.hass.data[DOMAIN][DATA_VEHICLE_INSTANCE]



import asyncio
import concurrent
import functools
import os
import time

import cv2
import numpy as np
from fhempy.lib.generic import FhemModule
from tflite_runtime.interpreter import Interpreter

from .. import fhem, utils


class object_detection(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.loop = asyncio.get_event_loop()
        self._cwd_path = os.getcwd()
        obj_det_mod_dir = os.path.dirname(os.path.abspath(__file__))
        self._labels_path = os.path.join(obj_det_mod_dir, "labelmap.txt")
        self._graph_path = os.path.join(obj_det_mod_dir, "detect.tflite")
        self._detection_task = None
        self._prev_objects = None
        self._stop_detection = False
        self._attr_detection_interval = 2
        self._attr_detection_threshold = 0.6
        self._attr_list = {
            "detection_interval": {"default": 2, "format": "float"},
            "detection_threshold": {"default": 0.6, "format": "float"},
        }
        self.set_attr_config(self._attr_list)
        set_list_conf = {"start": {}, "detect_once": {}, "stop": {}}
        self.set_set_config(set_list_conf)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 5:
            return "Usage: define obj_detection fhempy object_detection <image/stream> <LOCATION>"
        if args[3] == "image":
            if args[4][0] == "/":
                self._source_uri = args[4]
            else:
                self._source_uri = os.path.join(self._cwd_path, args[4])
        else:
            self._source_uri = args[4]
        self._source_type = args[3]

        self.logger.debug(f"Source URI: {self._source_uri}")
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    def run_stream_object_detection(self):
        # Load the label map
        with open(self._labels_path, "r") as f:
            labels = [line.strip().replace(" ", "_") for line in f.readlines()]

        # Have to do a weird fix for label map if using the COCO "starter model" from
        # https://www.tensorflow.org/lite/models/object_detection/overview
        # First label is '???', which has to be removed.
        if labels[0] == "???":
            del labels[0]
        interpreter = Interpreter(model_path=self._graph_path)
        interpreter.allocate_tensors()

        # Get model details
        input_details = interpreter.get_input_details()
        output_details = interpreter.get_output_details()
        height = input_details[0]["shape"][1]
        width = input_details[0]["shape"][2]
        floating_model = input_details[0]["dtype"] == np.float32
        input_mean = 127.5
        input_std = 127.5

        # Initialize frame rate calculation
        frame_rate_calc = 1
        freq = cv2.getTickFrequency()

        # resolution
        imW = 1280
        imH = 720

        # Initialize video stream
        videostream = VideoStream(
            self._source_uri, resolution=(imW, imH), framerate=30, loop=self.loop
        ).start()
        time.sleep(1)

        # for frame1 in camera.capture_continuous(rawCapture, format="bgr",use_video_port=True):
        while True:

            try:
                # Start timer (for calculating frame rate)
                # t1 = cv2.getTickCount()

                # Grab frame from video stream
                frame1 = videostream.read()
                if frame1 is None:
                    continue

                # Acquire frame and resize to expected shape [1xHxWx3]
                frame = frame1.copy()
                frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                frame_resized = cv2.resize(frame_rgb, (width, height))
                input_data = np.expand_dims(frame_resized, axis=0)

                # Normalize pixel values if using a floating model (i.e. if model is non-quantized)
                if floating_model:
                    input_data = (np.float32(input_data) - input_mean) / input_std

                # Perform the actual detection by running the model with the image as input
                interpreter.set_tensor(input_details[0]["index"], input_data)
                interpreter.invoke()

                # Retrieve detection results
                boxes = interpreter.get_tensor(output_details[0]["index"])[
                    0
                ]  # Bounding box coordinates of detected objects
                classes = interpreter.get_tensor(output_details[1]["index"])[
                    0
                ]  # Class index of detected objects
                scores = interpreter.get_tensor(output_details[2]["index"])[
                    0
                ]  # Confidence of detected objects
                # num = interpreter.get_tensor(output_details[3]['index'])[0]  # Total number of detected objects (inaccurate and not needed)

                detected_objects = []
                # Loop over all detections and draw detection box if confidence is above minimum threshold
                for i in range(len(scores)):
                    if (scores[i] > self._attr_detection_threshold) and (
                        scores[i] <= 1.0
                    ):

                        # Get bounding box coordinates and draw box
                        # Interpreter can return coordinates that are outside of image dimensions, need to force them to be within image using max() and min()
                        # ymin = int(max(1,(boxes[i][0] * imH)))
                        # xmin = int(max(1,(boxes[i][1] * imW)))
                        # ymax = int(min(imH,(boxes[i][2] * imH)))
                        # xmax = int(min(imW,(boxes[i][3] * imW)))

                        # cv2.rectangle(frame, (xmin,ymin), (xmax,ymax), (10, 255, 0), 2)

                        # Draw label
                        object_name = labels[
                            int(classes[i])
                        ]  # Look up object name from "labels" array using class index
                        label = "%s: %d%%" % (
                            object_name,
                            int(scores[i] * 100),
                        )  # Example: 'person: 72%'
                        # labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2) # Get font size
                        # label_ymin = max(ymin, labelSize[1] + 10) # Make sure not to draw label too close to top of window
                        # cv2.rectangle(frame, (xmin, label_ymin-labelSize[1]-10), (xmin+labelSize[0], label_ymin+baseLine-10), (255, 255, 255), cv2.FILLED) # Draw white box to put label text in
                        # cv2.putText(frame, label, (xmin, label_ymin-7), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2) # Draw label text

                        detected_objects.append(
                            {"object": object_name, "score": int(scores[i] * 100)}
                        )

                asyncio.run_coroutine_threadsafe(
                    self.update_readings(detected_objects), self.loop
                ).result()

                if self._stop_detection:
                    asyncio.run_coroutine_threadsafe(
                        fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1),
                        self.loop,
                    ).result()
                    videostream.stop()
                    return

                # Draw framerate in corner of frame
                # cv2.putText(frame,'FPS: {0:.2f}'.format(frame_rate_calc),(30,50),cv2.FONT_HERSHEY_SIMPLEX,1,(255,255,0),2,cv2.LINE_AA)

                # All the results have been drawn on the frame, so it's time to display it.
                # cv2.imshow('Object detector', frame)

                # Calculate framerate
                # t2 = cv2.getTickCount()
                # time1 = (t2-t1)/freq
                # frame_rate_calc= 1/time1
                time.sleep(self._attr_detection_interval)
            except asyncio.CancelledError:
                videostream.stop()
                return

    def run_image_object_detection(self):
        detected_objects = []
        # Load the label map
        with open(self._labels_path, "r") as f:
            labels = [line.strip() for line in f.readlines()]

        # Have to do a weird fix for label map if using the COCO "starter model" from
        # https://www.tensorflow.org/lite/models/object_detection/overview
        # First label is '???', which has to be removed.
        if labels[0] == "???":
            del labels[0]
        interpreter = Interpreter(model_path=self._graph_path)
        interpreter.allocate_tensors()

        # Get model details
        input_details = interpreter.get_input_details()
        output_details = interpreter.get_output_details()
        height = input_details[0]["shape"][1]
        width = input_details[0]["shape"][2]
        floating_model = input_details[0]["dtype"] == np.float32
        input_mean = 127.5
        input_std = 127.5

        # Load image and resize to expected shape [1xHxWx3]
        image_path = self._source_uri
        image = cv2.imread(image_path)
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        imH, imW, _ = image.shape
        image_resized = cv2.resize(image_rgb, (width, height))
        input_data = np.expand_dims(image_resized, axis=0)

        # Normalize pixel values if using a floating model (i.e. if model is non-quantized)
        if floating_model:
            input_data = (np.float32(input_data) - input_mean) / input_std

        # Perform the actual detection by running the model with the image as input
        interpreter.set_tensor(input_details[0]["index"], input_data)
        interpreter.invoke()

        # Retrieve detection results
        boxes = interpreter.get_tensor(output_details[0]["index"])[
            0
        ]  # Bounding box coordinates of detected objects
        classes = interpreter.get_tensor(output_details[1]["index"])[
            0
        ]  # Class index of detected objects
        scores = interpreter.get_tensor(output_details[2]["index"])[
            0
        ]  # Confidence of detected objects
        # num = interpreter.get_tensor(output_details[3]['index'])[0]  # Total number of detected objects (inaccurate and not needed)

        # Loop over all detections and draw detection box if confidence is above minimum threshold
        for i in range(len(scores)):
            if (scores[i] > self._attr_detection_threshold) and (scores[i] <= 1.0):

                # Get bounding box coordinates and draw box
                # Interpreter can return coordinates that are outside of image dimensions, need to force them to be within image using max() and min()
                ymin = int(max(1, (boxes[i][0] * imH)))
                xmin = int(max(1, (boxes[i][1] * imW)))
                ymax = int(min(imH, (boxes[i][2] * imH)))
                xmax = int(min(imW, (boxes[i][3] * imW)))

                cv2.rectangle(image, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)

                object_name = labels[
                    int(classes[i])
                ]  # Look up object name from "labels" array using class index
                detected_objects.append(
                    {"object": object_name, "score": int(scores[i] * 100)}
                )

                # Draw label
                label = "%s: %d%%" % (
                    object_name,
                    int(scores[i] * 100),
                )  # Example: 'person: 72%'
                labelSize, baseLine = cv2.getTextSize(
                    label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2
                )  # Get font size
                label_ymin = max(
                    ymin, labelSize[1] + 10
                )  # Make sure not to draw label too close to top of window
                cv2.rectangle(
                    image,
                    (xmin, label_ymin - labelSize[1] - 10),
                    (xmin + labelSize[0], label_ymin + baseLine - 10),
                    (255, 255, 255),
                    cv2.FILLED,
                )  # Draw white box to put label text in
                cv2.putText(
                    image,
                    label,
                    (xmin, label_ymin - 7),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.7,
                    (0, 0, 0),
                    2,
                )  # Draw label text

                # output file with object rectangle and text
                # cv2.imwrite(self._output_file, image)
        return detected_objects

    async def set_start(self, hash, params):
        self._stop_detection = False
        await self.start_detection()

    async def set_detect_once(self, hash, params):
        self._stop_detection = True
        await self.start_detection()

    async def start_detection(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)
        if self._source_type == "image":
            self.create_async_task(self.image_detect_objects_loop())
        else:
            self._detection_task = utils.run_blocking_task(
                functools.partial(self.run_stream_object_detection)
            )
        return ""

    async def set_stop(self, hash, params):
        self._stop_detection = True

    async def image_detect_objects_loop(self):
        while True:
            await self.image_detect_objects()
            if self._stop_detection:
                break
            await asyncio.sleep(self._attr_detection_interval)
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    async def image_detect_objects(self):
        detected_objects = await utils.run_blocking(
            functools.partial(self.run_image_object_detection)
        )
        await self.update_readings(detected_objects)

    async def update_readings(self, detected_objects):
        try:
            all_objects = {}
            curr_objects = {}
            if self._prev_objects is None:
                await fhem.CommandDeleteReading(
                    self.hash, self.hash["NAME"] + " object_.*"
                )

            await fhem.readingsBeginUpdate(self.hash)
            for obj in detected_objects:
                obj_name = obj["object"]
                obj_score = obj["score"]
                if obj_name not in all_objects:
                    all_objects[obj_name] = 0
                all_objects[obj_name] += 1
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "object_" + obj_name, obj_score
                )
            for obj_name in all_objects:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "object_count_" + obj_name, all_objects[obj_name]
                )
                curr_objects[obj_name] = all_objects[obj_name]

            # set readings to 0 for objects which were not found
            if self._prev_objects is None:
                set_0_readings = {}
            else:
                set_0_readings = set(self._prev_objects) - set(curr_objects)
            for set_0_reading in set_0_readings:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "object_" + set_0_reading, 0
                )
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "object_count_" + set_0_reading, 0
                )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "objects_detected", ",".join(set(all_objects))
            )
            await fhem.readingsEndUpdate(self.hash, 1)

            self._prev_objects = curr_objects
        except Exception:
            self.logger.exception("Failed to update readings")


# Define VideoStream class to handle streaming of video from webcam in separate processing thread
# Source - Adrian Rosebrock, PyImageSearch: https://www.pyimagesearch.com/2015/12/28/increasing-raspberry-pi-fps-with-python-and-opencv/
class VideoStream:
    """Camera object that controls video streaming"""

    def __init__(self, source_uri, resolution=(640, 480), framerate=30, loop=None):
        self.loop = loop
        # Initialize the PiCamera and the camera image stream
        self.stream = cv2.VideoCapture(source_uri)
        ret = self.stream.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"MJPG"))
        ret = self.stream.set(cv2.CAP_PROP_FPS, framerate)
        ret = self.stream.set(3, resolution[0])
        ret = self.stream.set(4, resolution[1])

        # Read first frame from the stream
        (self.grabbed, self.frame) = self.stream.read()

        # Variable to control when the camera is stopped
        self.stopped = False

    def start(self):
        # Start the thread that reads frames from the video stream
        self._task = self.loop.create_task(self.update_task())
        return self

    async def update_task(self):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await self.loop.run_in_executor(pool, functools.partial(self.update))

    def update(self):
        # Keep looping indefinitely until the thread is stopped
        while True:
            try:
                # If the camera is stopped, stop the thread
                if self.stopped:
                    # Close camera resources
                    self.stream.release()
                    return

                # Otherwise, grab the next frame from the stream
                (self.grabbed, self.frame) = self.stream.read()
            except asyncio.CancelledError:
                self.stream.release()
                return

    def read(self):
        # Return the most recent frame
        return self.frame

    def stop(self):
        # Indicate that the camera and thread should be stopped
        self.stopped = True
